import 'package:flutter/material.dart';
//import 'package:flutter_pagewise/flutter_pagewise.dart';
import 'package:adeptarch/src/pagination/flutter_pagewise.dart';//this implementation fixes a bug in resetting and fetching data too fast
import 'package:adeptarch/src/pagination/request_paginate.dart';
import 'package:adeptutils/adeptutils.dart';
 import 'package:flutter/foundation.dart';
import 'package:adeptarch/src/resource/resource.dart';
import 'dart:async';

typedef Future PageRequest(int pageIndex);
typedef void ErrorNotifier(context, Resource resourceError);

class PaginateController<T> extends PagewiseLoadController<T>{

  String defaultTextNoData;

  PaginateController({
    PageFuture<T> pageControl,
    PageRequest pageRequest,
    int pageSize = RequestPaginate.DEFAULT_SIZE,
    this.defaultTextNoData = "No data to show",
  })
      :
  super(
    pageFuture: pageControl ?? (index) async{
      try{
        Object response;
        List<T> result;
        if(pageRequest != null) response = await pageRequest(index);
        if(response is Resource) result = _checkResource(response, defaultTextNoData);
        if(response is List) result = response;
        if(result == null) result = [];
        if(index != null && index == 0 && result.isEmpty)
          throw Resource.error(null, defaultTextNoData);
        return result;
      }catch(e){
        return Future.error(e);
      }
    },
    pageSize: pageSize
  );

  static List _checkResource(Resource<List> resource, String defaultTextNoData){
    if(resource == null) return null;
    switch(resource.status){
      case ResourceStatus.SUCCESS:
        if(resource.exists())
          return resource.data;
        else{
          resource.status = ResourceStatus.ERROR;
          resource.message = defaultTextNoData;
          throw resource;
        }
        break;
      case ResourceStatus.TRANSIENT:
      case ResourceStatus.LOADING:
        return [];
      case ResourceStatus.ERROR:
      default:
        throw resource;
    }
  }

}

typedef String LocalizedMessageFromException(dynamic exception);
typedef String LocalizedMessageFromHttpResultCode(int code);
typedef Widget ErrorViewBuilder(BuildContext context, String message, PagewiseLoadController pageLoadController);

class PaginatedListView<T> extends PagewiseListView<T> {

  PaginatedListView({
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode getLocalizedMessageFromHttpResultCode,
    Key key,
    EdgeInsetsGeometry padding,
    bool primary,
    bool addSemanticIndexes = true,
    int semanticChildCount,
    bool shrinkWrap: false,
    ScrollController controller,
    PagewiseLoadController<T> pageLoadController,
    double itemExtent,
    bool addAutomaticKeepAlives: true,
    Axis scrollDirection: Axis.vertical,
    bool addRepaintBoundaries: true,
    double cacheExtent,
    ScrollPhysics physics = const AlwaysScrollableScrollPhysics(),
    bool reverse: false,
    int pageSize,
    PageFuture<T> pageFuture,
    LoadingBuilder loadingBuilder,
    RetryBuilder retryBuilder,
    NoItemsFoundBuilder noItemsFoundBuilder,
    bool showRetry: false,
    @required ItemBuilder<T> itemBuilder,
    ErrorNotifier errorNotifier,
    ErrorViewBuilder errorViewBuilder,
    ErrorBuilder errorBuilder,
  }) : super(
    key: key,
    padding: padding,
    primary: primary,
    addSemanticIndexes: addSemanticIndexes,
    semanticChildCount: semanticChildCount,
    shrinkWrap: shrinkWrap,
    controller: controller,
    pageLoadController: pageLoadController,
    itemExtent: itemExtent,
    addAutomaticKeepAlives: addAutomaticKeepAlives,
    scrollDirection: scrollDirection,
    addRepaintBoundaries: addRepaintBoundaries,
    cacheExtent: cacheExtent,
    physics: physics,
    reverse: reverse,
    pageSize: pageSize,
    pageFuture: pageFuture,
    loadingBuilder: loadingBuilder ?? (context) => RefreshProgressIndicator(),
    retryBuilder: retryBuilder,
    noItemsFoundBuilder: noItemsFoundBuilder,
    showRetry: showRetry,
    itemBuilder: itemBuilder,
    errorBuilder: errorBuilder ?? (context, error){
      Resource resourceError = Resource.error(null, defaultTextError);
      try{
        if(error is Exception) {
          Exception exc = ExceptionUtils.parseException(error);
          String message = (getLocalizedMessageFromException != null ?
                              getLocalizedMessageFromException(exc) : null) ?? exc.toString();
          resourceError = Resource.error(null, message, exception: exc);
        } else
        if(error is String) resourceError = Resource.error(null, error);
        if(error is Resource) resourceError = error;

        if(TextUtils.isEmpty(resourceError.message)){
          if(resourceError.exception != null){
            Exception exc = ExceptionUtils.parseException(error);
            String message = (getLocalizedMessageFromException != null ?
            getLocalizedMessageFromException(exc) : null) ?? exc.toString();
            resourceError.message = message;
          } else
          if(resourceError.code != null){
            resourceError.message = (getLocalizedMessageFromHttpResultCode != null ?
            getLocalizedMessageFromHttpResultCode(resourceError.code) : null);
          }
        }

        if(TextUtils.isEmpty(resourceError.message)) resourceError.message = defaultTextError;

      }catch(e){
        Exception exc = ExceptionUtils.parseException(error);
        String message = (getLocalizedMessageFromException != null ?
        getLocalizedMessageFromException(exc) : null) ?? exc.toString();
        resourceError = Resource.error(null, message, exception: exc);
      }

      if(errorNotifier != null) errorNotifier(context, resourceError);
      if(errorViewBuilder != null) return errorViewBuilder(context, resourceError.message, pageLoadController);
      return Container(
        child: Column(
          children: <Widget>[
            Container(
              alignment: Alignment.center,
              padding: EdgeInsets.zero,
              child: Column(
                crossAxisAlignment: CrossAxisAlignment.center,
                mainAxisAlignment: MainAxisAlignment.center,
                children: <Widget>[
                  Icon(Icons.error_outline, size: 32, color: Colors.black54,),
                  Divider(color: Colors.transparent, height: 4,),
                  Center(
                    child: Text("${resourceError.message}", style: TextStyle(color: Colors.black54, fontSize: 14), textAlign: TextAlign.center,)
                  ),
                ],
              ),
            ),
            FloatingActionButton(
              child: Icon(Icons.refresh, size: 24, color: Colors.white,),
              mini: true,
              onPressed: () {
                if(pageLoadController != null) pageLoadController.retry();
              }
            ),
          ],
        ),
      );
    },
  );
}

class PaginatedGridView<T> extends PagewiseGridView<T>{

  PaginatedGridView.count({
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode getLocalizedMessageFromHttpResultCode,
    Key key,
    EdgeInsetsGeometry padding,
    @required int crossAxisCount,
    double childAspectRatio = 1.0,
    double crossAxisSpacing = 0.0,
    double mainAxisSpacing = 0.0,
    bool addSemanticIndexes = true,
    int semanticChildCount,
    bool primary,
    bool shrinkWrap: false,
    ScrollController controller,
    PagewiseLoadController<T> pageLoadController,
    bool addAutomaticKeepAlives: true,
    Axis scrollDirection: Axis.vertical,
    bool addRepaintBoundaries: true,
    double cacheExtent,
    ScrollPhysics physics = const AlwaysScrollableScrollPhysics(),
    bool reverse: false,
    int pageSize,
    PageFuture<T> pageFuture,
    LoadingBuilder loadingBuilder,
    RetryBuilder retryBuilder,
    NoItemsFoundBuilder noItemsFoundBuilder,
    bool showRetry: false,
    @required ItemBuilder<T> itemBuilder,
    ErrorNotifier errorNotifier,
    ErrorViewBuilder errorViewBuilder,
    ErrorBuilder errorBuilder
  }) : super.count(
    key: key,
    padding: padding,
    crossAxisCount: crossAxisCount,
    childAspectRatio: childAspectRatio,
    crossAxisSpacing: crossAxisSpacing,
    mainAxisSpacing: mainAxisSpacing,
    primary: primary,
    addSemanticIndexes: addSemanticIndexes,
    semanticChildCount: semanticChildCount,
    shrinkWrap: shrinkWrap,
    controller: controller,
    pageLoadController: pageLoadController,
    addAutomaticKeepAlives: addAutomaticKeepAlives,
    scrollDirection: scrollDirection,
    addRepaintBoundaries: addRepaintBoundaries,
    cacheExtent: cacheExtent,
    physics: physics,
    reverse: reverse,
    pageSize: pageSize,
    pageFuture: pageFuture,
    loadingBuilder: loadingBuilder ?? (context) => RefreshProgressIndicator(),
    retryBuilder: retryBuilder,
    noItemsFoundBuilder: noItemsFoundBuilder,
    showRetry: showRetry,
    itemBuilder: itemBuilder,
    errorBuilder: errorBuilder ?? (context, error){
      Resource resourceError = Resource.error(null, defaultTextError);
      try{
        if(error is Exception) {
          Exception exc = ExceptionUtils.parseException(error);
          String message = (getLocalizedMessageFromException != null ?
          getLocalizedMessageFromException(exc) : null) ?? exc.toString();
          resourceError = Resource.error(null, message, exception: exc);
        } else
        if(error is String) resourceError = Resource.error(null, error);
        if(error is Resource) resourceError = error;

        if(TextUtils.isEmpty(resourceError.message)){
          if(resourceError.exception != null){
            Exception exc = ExceptionUtils.parseException(error);
            String message = (getLocalizedMessageFromException != null ?
            getLocalizedMessageFromException(exc) : null) ?? exc.toString();
            resourceError.message = message;
          } else
          if(resourceError.code != null){
            resourceError.message = (getLocalizedMessageFromHttpResultCode != null ?
            getLocalizedMessageFromHttpResultCode(resourceError.code) : null);
          }
        }

        if(TextUtils.isEmpty(resourceError.message)) resourceError.message = defaultTextError;

      }catch(e){
        Exception exc = ExceptionUtils.parseException(error);
        String message = (getLocalizedMessageFromException != null ?
        getLocalizedMessageFromException(exc) : null) ?? exc.toString();
        resourceError = Resource.error(null, message, exception: exc);
      }

      if(errorNotifier != null) errorNotifier(context, resourceError);
      if(errorViewBuilder != null) return errorViewBuilder(context, resourceError.message, pageLoadController);
      return Container(
        child: Column(
          children: <Widget>[
            Container(
              alignment: Alignment.center,
              padding: EdgeInsets.zero,
              child: Column(
                crossAxisAlignment: CrossAxisAlignment.center,
                mainAxisAlignment: MainAxisAlignment.center,
                children: <Widget>[
                  Icon(Icons.error_outline, size: 32, color: Colors.black54,),
                  Divider(color: Colors.transparent, height: 4,),
                  Center(
                      child: Text("${resourceError.message}", style: TextStyle(color: Colors.black54, fontSize: 14), textAlign: TextAlign.center,)
                  ),
                ],
              ),
            ),
            FloatingActionButton(
                child: Icon(Icons.refresh, size: 24, color: Colors.white,),
                mini: true,
                onPressed: () {
                  if(pageLoadController != null) pageLoadController.retry();
                }
            ),
          ],
        ),
      );
    },
  );
}